from braintree.graphql.enums import (
    RecommendedPaymentOption,
    Recommendations
)
from braintree.graphql.inputs import (
    PhoneInput,
    CustomerSessionInput,
    CreateCustomerSessionInput,
    UpdateCustomerSessionInput,
    CustomerRecommendationsInput,
)
from braintree.graphql.types import (
    CustomerRecommendationsPayload,
    PaymentOptions
)
from braintree.graphql.unions import (
    CustomerRecommendations
)




import warnings
from decimal import Decimal
from braintree.attribute_getter import AttributeGetter
from braintree.transaction_details import TransactionDetails
from braintree.dispute_details import DisputeEvidence, DisputeStatusHistory, DisputePayPalMessage
from braintree.configuration import Configuration

class Dispute(AttributeGetter):
    # NEXT_MAJOR_VERSION this can be an enum! they were added as of python 3.4 and we support 3.5+
    class Status(object):
        """
        Constants representing dispute statuses. Available types are:

        * braintree.Dispute.Status.Accepted
        * braintree.Dispute.Status.AutoAccepted
        * braintree.Dispute.Status.Disputed
        * braintree.Dispute.Status.Expired
        * braintree.Dispute.Status.Lost
        * braintree.Dispute.Status.Open
        * braintree.Dispute.Status.UnderReview
        * braintree.Dispute.Status.Won
        """
        Accepted = "accepted"
        AutoAccepted = "auto_accepted"
        Disputed = "disputed"
        Expired = "expired"
        Lost = "lost"
        Open  = "open"
        UnderReview = "under_review"
        Won  = "won"

    # NEXT_MAJOR_VERSION this can be an enum! they were added as of python 3.4 and we support 3.5+
    class Reason(object):
        """
        Constants representing dispute reasons. Available types are:

        * braintree.Dispute.Reason.CancelledRecurringTransaction
        * braintree.Dispute.Reason.CreditNotProcessed
        * braintree.Dispute.Reason.Duplicate
        * braintree.Dispute.Reason.Fraud
        * braintree.Dispute.Reason.General
        * braintree.Dispute.Reason.InvalidAccount
        * braintree.Dispute.Reason.NotRecognized
        * braintree.Dispute.Reason.ProductNotReceived
        * braintree.Dispute.Reason.ProductUnsatisfactory
        * braintree.Dispute.Reason.Retrieval
        * braintree.Dispute.Reason.TransactionAmountDiffers
        """
        CancelledRecurringTransaction = "cancelled_recurring_transaction"
        CreditNotProcessed            = "credit_not_processed"
        Duplicate                     = "duplicate"
        Fraud                         = "fraud"
        General                       = "general"
        InvalidAccount                = "invalid_account"
        NotRecognized                 = "not_recognized"
        ProductNotReceived            = "product_not_received"
        ProductUnsatisfactory         = "product_unsatisfactory"
        Retrieval                     = "retrieval"
        TransactionAmountDiffers      = "transaction_amount_differs"

    # NEXT_MAJOR_VERSION this can be an enum! they were added as of python 3.4 and we support 3.5+
    class Kind(object):
        """
        Constants representing dispute kinds. Available types are:

        * braintree.Dispute.Kind.Chargeback
        * braintree.Dispute.Kind.PreArbitration
        * braintree.Dispute.Kind.Retrieval
        """
        Chargeback     = "chargeback"
        PreArbitration = "pre_arbitration"
        Retrieval      = "retrieval"

    # NEXT_MAJOR_VERSION Remove this enum
    class ChargebackProtectionLevel(object):
        """
        Constants representing dispute ChargebackProtectionLevel. Available types are:

        * braintree.Dispute.ChargebackProtectionLevel.EFFORTLESS
        * braintree.Dispute.ChargebackProtectionLevel.STANDARD
        * braintree.Dispute.ChargebackProtectionLevel.NOT_PROTECTED
        """
        warnings.warn("Use ProtectionLevel enum instead", DeprecationWarning)
        Effortless     = "effortless"
        Standard       = "standard"
        NotProtected   = "not_protected"

    # NEXT_MAJOR_VERSION this can be an enum! they were added as of python 3.4 and we support 3.5+
    class PreDisputeProgram(object):
        """
        Constants representing dispute pre-dispute programs. Available types are:

        * braintree.Dispute.PreDisputeProgram.NONE
        * braintree.Dispute.PreDisputeProgram.VisaRdr
        """
        NONE = "none"
        VisaRdr = "visa_rdr"

    # NEXT_MAJOR_VERSION this can be an enum! they were added as of python 3.4 and we support 3.5+
    class ProtectionLevel(object):
        """
        Constants representing dispute ProtectionLevel. Available types are:

        * braintree.Dispute.ProtectionLevel.EffortlessCBP
        * braintree.Dispute.ProtectionLevel.StandardCBP
        * braintree.Dispute.ProtectionLevel.NoProtection
        """
        EffortlessCBP  = "Effortless Chargeback Protection tool"
        StandardCBP    = "Chargeback Protection tool"
        NoProtection   = "No Protection"

    @staticmethod
    def accept(id):
        """
        Accept a dispute, given a dispute_id.
        This will raise a :class:`NotFoundError <braintree.exceptions.not_found_error.NotFoundError>` if the provided dispute_id
        is not found. ::

            result = braintree.Dispute.accept("my_dispute_id")
        """

        return Configuration.gateway().dispute.accept(id)

    @staticmethod
    def add_file_evidence(dispute_id, document_upload_id):
        """
        Adds file evidence to a dispute, given a dispute_id and a document_upload_id.

        This will raise a :class:`NotFoundError <braintree.exceptions.not_found_error.NotFoundError>` if the provided dispute_id
        is not found. ::

            document = braintree.DocumentUpload.create({
                "kind": braintree.DocumentUpload.Kind.EvidenceDocument,
                "file": open("/path/to/evidence.pdf", "rb")
            })

            result = braintree.Dispute.add_file_evidence("my_dispute_id", document.id)
        """

        return Configuration.gateway().dispute.add_file_evidence(dispute_id, document_upload_id)

    @staticmethod
    def add_text_evidence(id, content_or_request):
        """
        Adds text evidence to a dispute, given a dispute_id.

        This will raise a :class:`NotFoundError <braintree.exceptions.not_found_error.NotFoundError>` if the provided dispute_id
        is not found. ::

            result = braintree.Dispute.add_text_evidence("my_dispute_id", "my_evidence")

            or

            result = braintree.Dispute.add_text_evidence("my_dispute_id", { "content": "UPS", "tag": "CARRIER_NAME", "sequence_number": "1" })
        """
        return Configuration.gateway().dispute.add_text_evidence(id, content_or_request)

    @staticmethod
    def finalize(id):
        """
        Finalize a dispute, given a dispute_id.
        This will raise a :class:`NotFoundError <braintree.exceptions.not_found_error.NotFoundError>` if the provided dispute_id
        is not found. ::

            result = braintree.Dispute.finalize("my_dispute_id")
        """

        return Configuration.gateway().dispute.finalize(id)

    @staticmethod
    def find(id):
        """
        Find an dispute, given a dispute_id.  This does not return a result
        object.  This will raise a :class:`NotFoundError <braintree.exceptions.not_found_error.NotFoundError>` if the provided dispute_id
        is not found. ::

            dispute = braintree.Dispute.find("my_dispute_id")
        """

        return Configuration.gateway().dispute.find(id)

    @staticmethod
    def remove_evidence(id, evidence_id):
        """
        Remove evidence on a dispute.
        This will raise a :class:`NotFoundError <braintree.exceptions.not_found_error.NotFoundError>` if the provided dispute_id or evidence_id
        is not found. ::

            result = braintree.Dispute.remove_evidence("my_dispute_id", "my_evidence_id")
        """

        return Configuration.gateway().dispute.remove_evidence(id, evidence_id)

    @staticmethod
    def search(*query):
        """
        Searches for disputes, given a DisputeSearch query.

            collection = braintree.Dispute.search([
                braintree.DisputeSearch.id == "the_dispute_id"
            ])

            for dispute in collection.items:
                print dispute.id
        """

        return Configuration.gateway().dispute.search(*query)

    def __init__(self, attributes):
        if "chargeback_protection_level" in attributes:
            warnings.warn("Use protection_level attribute instead", DeprecationWarning)
        AttributeGetter.__init__(self, attributes)

        if "amount" in attributes and getattr(self, "amount", None) is not None:
            self.amount = Decimal(self.amount)
        if "amount_disputed" in attributes and getattr(self, "amount_disputed", None) is not None:
            self.amount_disputed = Decimal(self.amount_disputed)
        if "amount_won" in attributes and getattr(self, "amount_won", None) is not None:
            self.amount_won = Decimal(self.amount_won)
        if "chargeback_protection_level" in attributes and getattr(self, "chargeback_protection_level", None) in [self.ChargebackProtectionLevel.Effortless, self.ChargebackProtectionLevel.Standard]:
            self.protection_level = eval("self.ProtectionLevel.{0}CBP".format(self.chargeback_protection_level.capitalize()))
        else:
            self.protection_level = self.ProtectionLevel.NoProtection
        if "transaction" in attributes:
            self.transaction_details = TransactionDetails(attributes.pop("transaction"))
            self.transaction = self.transaction_details
        if "evidence" in attributes and getattr(self, "evidence", None) is not None:
            self.evidence = [DisputeEvidence(evidence) for evidence in self.evidence]
        if "paypal_messages" in attributes and getattr(self, "paypal_messages", None) is not None:
            self.paypal_messages = [DisputePayPalMessage(paypal_message) for paypal_message in self.paypal_messages]
        if "status_history" in attributes and getattr(self, "status_history", None) is not None:
            self.status_history = [DisputeStatusHistory(status_history) for status_history in self.status_history]
        if "processor_comments" in attributes and self.processor_comments is not None:
            self.forwarded_comments = self.processor_comments



import braintree
from braintree.us_bank_account import UsBankAccount
from braintree.exceptions.not_found_error import NotFoundError

class UsBankAccountGateway(object):
    def __init__(self, gateway):
        self.gateway = gateway
        self.config = gateway.config

    def find(self, us_bank_account_token):
        try:
            if us_bank_account_token is None or us_bank_account_token.strip() == "":
                raise NotFoundError()

            response = self.config.http().get(self.config.base_merchant_path() + "/payment_methods/us_bank_account/" + us_bank_account_token)
            if "us_bank_account" in response:
                return UsBankAccount(self.gateway, response["us_bank_account"])
        except NotFoundError:
            raise NotFoundError("US bank account with token " + repr(us_bank_account_token) + " not found")




from braintree.exceptions.braintree_error import BraintreeError

class NotFoundError(BraintreeError):
    """
    Raised when an object is not found in the gateway, such as a Transaction.find("bad_id").

    See https://developer.paypal.com/braintree/docs/reference/general/exceptions/python#not-found-error
    """
    pass



from braintree.attribute_getter import AttributeGetter
from warnings import warn

class PackageDetails(AttributeGetter):
    """
    A class representing the package tracking information of a transaction.

    An example of package details including all available fields::

        result = braintree.PackageDetails.create({
            "id": "my_id",
            "carrier": "a_carrier",
            "tracking_number": "my_tracking_number",
            "paypal_tracking_id": "my_paypal_tracking_id",
            "paypal_tracker_id": "my_paypal_tracker_id",
        })

    """
    detail_list = [
        "id",
        "carrier",
        "tracking_number",
        # NEXT_MAJOR_VERSION remove paypal_tracking_id
        "paypal_tracking_id",
        "paypal_tracker_id",
    ]

    def __init__(self, attributes):
        AttributeGetter.__init__(self, attributes)



import braintree
from braintree.resource import Resource
from braintree.configuration import Configuration


class PayPalAccount(Resource):
    @staticmethod
    def find(paypal_account_token):
        return Configuration.gateway().paypal_account.find(paypal_account_token)

    @staticmethod
    def delete(paypal_account_token):
        return Configuration.gateway().paypal_account.delete(paypal_account_token)

    @staticmethod
    def update(paypal_account_token, params=None):
        if params is None:
            params = {}
        return Configuration.gateway().paypal_account.update(paypal_account_token, params)

    @staticmethod
    def signature():
        signature = [
            "token",
            {"options": ["make_default"]}
        ]
        return signature

    def __init__(self, gateway, attributes):
        Resource.__init__(self, gateway, attributes)
        if "subscriptions" in attributes:
            self.subscriptions = [braintree.subscription.Subscription(gateway, subscription) for subscription in self.subscriptions]



from base64 import b64decode, encodebytes
from contextlib import contextmanager
from datetime import date, datetime, timedelta
from decimal import Decimal
from enum import Enum
from http.client import HTTPConnection
from subprocess import Popen, PIPE
from urllib.parse import urlencode, quote_plus
import json
import os
import random
import re
import requests
import subprocess
import sys
import time
import unittest
import warnings

from braintree import *
from braintree.exceptions import *
from braintree.test.credit_card_numbers import CreditCardNumbers
from braintree.test.nonces import Nonces
from braintree.testing_gateway import *
from braintree.util import *

def reset_braintree_configuration():
    Configuration.configure(
        Environment.Development,
        "integration_merchant_id",
        "integration_public_key",
        "integration_private_key"
    )
reset_braintree_configuration()

class AdvancedFraudKountIntegrationMerchant:
    def __enter__(self):
        Configuration.configure(
            Environment.Development,
            "advanced_fraud_integration_merchant_id",
            "advanced_fraud_integration_public_key",
            "advanced_fraud_integration_private_key"
        )

    def __exit__(self, type, value, trace):
        reset_braintree_configuration()

class FraudProtectionEnterpriseIntegrationMerchant:
    def __enter__(self):
        Configuration.configure(
            Environment.Development,
            "fraud_protection_enterprise_integration_merchant_id",
            "fraud_protection_enterprise_integration_public_key",
            "fraud_protection_enterprise_integration_private_key"
        )

    def __exit__(self, type, value, trace):
        reset_braintree_configuration()

class EffortlessChargebackProtectionMerchant:
    def __enter__(self):
        Configuration.configure(
            Environment.Development,
            "fraud_protection_effortless_chargeback_protection_merchant_id",
            "effortless_chargeback_protection_public_key",
            "effortless_chargeback_protection_private_key"
        )

    def __exit__(self, type, value, trace):
        reset_braintree_configuration()

class DuplicateCheckingMerchant:
    def __enter__(self):
        Configuration.configure(
            Environment.Development,
            "dup_checking_integration_merchant_id",
            "dup_checking_integration_public_key",
            "dup_checking_integration_private_key"
        )

    def __exit__(self, type, value, trace):
        reset_braintree_configuration()

def showwarning(*_):
    pass
warnings.showwarning = showwarning

class TestHelper(object):
    adyen_merchant_account_id = "adyen_ma"
    aib_swe_ma_merchant_account_id = "aib_swe_ma"
    another_us_bank_merchant_account_id = "another_us_bank_merchant_account"
    card_processor_brl_merchant_account_id = "card_processor_brl"
    default_merchant_account_id = "sandbox_credit_card"
    fake_amex_direct_merchant_account_id = "fake_amex_direct_usd"
    fake_first_data_merchant_account_id = "fake_first_data_merchant_account"
    fake_venmo_account_merchant_account_id = "fake_first_data_venmo_account"
    hiper_brl_merchant_account_id = "hiper_brl"
    non_default_merchant_account_id = "sandbox_credit_card_non_default"
    non_default_sub_merchant_account_id = "sandbox_sub_merchant_account"
    pinless_debit_merchant_account_id = "pinless_debit"
    three_d_secure_merchant_account_id = "three_d_secure_merchant_account"
    us_bank_merchant_account_id = "us_bank_merchant_account"

    add_on_discount_plan = {
         "description": "Plan for integration tests -- with add-ons and discounts",
         "id": "integration_plan_with_add_ons_and_discounts",
         "price": Decimal("9.99"),
         "trial_duration": 2,
         "trial_duration_unit": Subscription.TrialDurationUnit.Day,
         "trial_period": True
    }

    billing_day_of_month_plan = {
         "description": "Plan for integration tests -- with billing day of month",
         "id": "integration_plan_with_billing_day_of_month",
         "billing_day_of_month": 5,
         "price": Decimal("8.88"),
    }

    trial_plan = {
        "description": "Plan for integration tests -- with trial",
        "id": "integration_trial_plan",
        "price": Decimal("43.21"),
        "trial_period": True,
        "trial_duration": 2,
        "trial_duration_unit": Subscription.TrialDurationUnit.Day
    }

    trialless_plan = {
        "description": "Plan for integration tests -- without a trial",
        "id": "integration_trialless_plan",
        "price": Decimal("12.34"),
        "trial_period": False
    }

    valid_token_characters = list("bcdfghjkmnpqrstvwxyz23456789")
    text_type = str
    raw_type = bytes

    @staticmethod
    def make_past_due(subscription, number_of_days_past_due=1):
        Configuration.gateway().testing.make_past_due(subscription.id, number_of_days_past_due)

    @staticmethod
    def escrow_transaction(transaction_id):
        Configuration.gateway().testing.escrow_transaction(transaction_id)

    @staticmethod
    def settle_transaction(transaction_id):
        return Configuration.gateway().testing.settle_transaction(transaction_id)

    @staticmethod
    def settlement_confirm_transaction(transaction_id):
        return Configuration.gateway().testing.settlement_confirm_transaction(transaction_id)

    @staticmethod
    def settlement_decline_transaction(transaction_id):
        return Configuration.gateway().testing.settlement_decline_transaction(transaction_id)

    @staticmethod
    def settlement_pending_transaction(transaction_id):
        return Configuration.gateway().testing.settlement_pending_transaction(transaction_id)

    @staticmethod
    def create_3ds_verification(merchant_account_id, params):
        return Configuration.gateway().testing.create_3ds_verification(merchant_account_id, params)

    @staticmethod
    @contextmanager
    def other_merchant(merchant_id, public_key, private_key):
        old_merchant_id = Configuration.merchant_id
        old_public_key = Configuration.public_key
        old_private_key = Configuration.private_key

        Configuration.merchant_id = merchant_id
        Configuration.public_key = public_key
        Configuration.private_key = private_key

        try:
            yield
        finally:
            Configuration.merchant_id = old_merchant_id
            Configuration.public_key = old_public_key
            Configuration.private_key = old_private_key

    @staticmethod
    def includes(collection, expected):
        for item in collection.items:
            if item.id == expected.id:
                return True
        return False

    @staticmethod
    def in_list(collection, expected):
        for item in collection:
            if item == expected:
                return True
        return False

    @staticmethod
    def includes_status(collection, status):
        for item in collection.items:
            if item.status == status:
                return True
        return False

    @staticmethod
    def now_minus_offset(offset):
        now = datetime.utcnow()
        return (now - timedelta(hours=offset)).strftime("%Y-%m-%d")

    @staticmethod
    def unique(some_list):
        return set(some_list)

    @staticmethod
    def __headers():
        return {
            "Accept": "application/xml",
            "Content-type": "application/x-www-form-urlencoded",
        }

    @staticmethod
    def generate_decoded_client_token(params=None):
        client_token = None
        if params:
            client_token = ClientToken.generate(params)
        else:
            client_token = ClientToken.generate()

        decoded_client_token = b64decode(client_token).decode()
        return decoded_client_token

    @staticmethod
    def nonce_for_paypal_account(paypal_account_details):
        client_token = json.loads(TestHelper.generate_decoded_client_token())
        client = ClientApiHttp(Configuration.instantiate(), {
            "authorization_fingerprint": client_token["authorizationFingerprint"]
        })

        _, nonce = client.get_paypal_nonce(paypal_account_details)
        return nonce

    @staticmethod
    def random_token_block(x):
        string = ""
        for i in range(6):
            string += random.choice(TestHelper.valid_token_characters)
        return string

    @staticmethod
    def generate_valid_us_bank_account_nonce(routing_number="021000021", account_number="567891234"):
        query = '''
          mutation TokenizeUsBankAccount($input: TokenizeUsBankAccountInput!) {
            tokenizeUsBankAccount(input: $input) {
              paymentMethod {
                id
              }
            }
          }
        '''

        variables = {
            "input": {
                "usBankAccount": {
                    "accountNumber": account_number,
                    "routingNumber": routing_number,
                    "accountType": "CHECKING",
                    "individualOwner": {
                        "firstName": "Dan",
                        "lastName": "Schulman"
                     },
                    "achMandate": "cl mandate text",
                    "billingAddress": {
                        "streetAddress": "123 Ave",
                        "state": "CA",
                        "city": "San Francisco",
                        "zipCode": "94112"
                     }
                }
            }
        }

        graphql_request = {
            "query": query,
            "variables": variables
        }

        response = TestHelper.__send_graphql_request(graphql_request)
        return response["data"]["tokenizeUsBankAccount"]["paymentMethod"]["id"]

    @staticmethod
    def generate_plaid_us_bank_account_nonce():
        query = '''
          mutation TokenizeUsBankLogin($input: TokenizeUsBankLoginInput!) {
            tokenizeUsBankLogin(input: $input) {
              paymentMethod {
                id
              }
            }
          }
        '''

        variables = {
            "input": {
                "usBankLogin": {
                    "publicToken": "good",
                    "accountId": "plaid_account_id",
                    "accountType": "CHECKING",
                    "businessOwner": {
                        "businessName": "PayPal, Inc."
                     },
                    "achMandate": "cl mandate text",
                    "billingAddress": {
                        "streetAddress": "123 Ave",
                        "state": "CA",
                        "city": "San Francisco",
                        "zipCode": "94112"
                     }
                }
            }
        }

        graphql_request = {
            "query": query,
            "variables": variables
        }

        response = TestHelper.__send_graphql_request(graphql_request)
        return response["data"]["tokenizeUsBankLogin"]["paymentMethod"]["id"]

    @staticmethod
    def generate_invalid_us_bank_account_nonce():
        token = "tokenusbankacct"
        for i in range(4):
            token += "_" + TestHelper.random_token_block('d')
        token += "_xxx"
        return token

    @staticmethod
    def generate_three_d_secure_nonce(gateway, params):
        url = gateway.config.base_merchant_path() + "/three_d_secure/create_nonce/" + TestHelper.three_d_secure_merchant_account_id
        response = gateway.config.http().post(url, params)
        return response["payment_method_nonce"]["nonce"]

    @staticmethod
    def create_disputed_transaction():
        if hasattr(TestHelper, 'disputed_transaction'):
            return TestHelper.disputed_transaction

        disputed_transaction = Transaction.sale({
            "amount": "10.00",
            "credit_card": {
                "number": CreditCardNumbers.Disputes.Chargeback,
                "expiration_date": "04/2018"
                }
            })

        for _ in range(1, 60):
            transactions = Transaction.search([
                TransactionSearch.id == disputed_transaction.transaction.id,
                TransactionSearch.dispute_date == datetime.today()
            ])

            if transactions.maximum_size == 1:
                TestHelper.disputed_transaction = transactions.first
                return TestHelper.disputed_transaction
            else:
                time.sleep(1)

        raise ValueError('Disputed transaction could not be found')

    @staticmethod
    def create_grant(gateway, params):
        config = gateway.config
        response = config.http().post("/oauth_testing/grants", {
            "grant": params
        })

        return response["grant"]["code"]

    @staticmethod
    def create_payment_method_grant_fixtures():
        config = Configuration(
            merchant_id="integration_merchant_public_id",
            public_key="oauth_app_partner_user_public_key",
            private_key="oauth_app_partner_user_private_key",
            environment=Environment.Development
        )

        gateway = BraintreeGateway(config)
        customer = gateway.customer.create().customer
        credit_card = gateway.credit_card.create(
            params={
                "customer_id": customer.id,
                "number": "4111111111111111",
                "expiration_date": "05/2009",
                "billing_address": {
                    "first_name": "Jon",
                    "last_name": "Doe",
                    "postal_code": "95131"
                }
            }
        ).credit_card

        oauth_app_gateway = BraintreeGateway(
            client_id="client_id$development$integration_client_id",
            client_secret="client_secret$development$integration_client_secret",
            environment=Environment.Development
        )
        code = TestHelper.create_grant(oauth_app_gateway, {
            "merchant_public_id": "integration_merchant_id",
            "scope": "grant_payment_method"
        })
        access_token = oauth_app_gateway.oauth.create_token_from_code({
            "code": code
        }).credentials.access_token

        granting_gateway = BraintreeGateway(
            access_token=access_token,
        )

        return (granting_gateway, credit_card)

    @staticmethod
    def sample_notification_from_xml(xml):
        gateway = Configuration.gateway()
        payload = encodebytes(xml)
        hmac_payload = Crypto.sha1_hmac_hash(gateway.config.private_key, payload)
        signature = "%s|%s" % (gateway.config.public_key, hmac_payload)
        return {'bt_signature': signature, 'bt_payload': payload}

    @staticmethod
    def __send_graphql_request(graphql_request):
        client_token = json.loads(TestHelper.generate_decoded_client_token())
        headers = {
            "Content-Type": "application/json",
            "Braintree-Version": "2016-10-07",
            "Authorization": "Bearer " + client_token["braintree_api"]["access_token"]
        }
        resp = requests.post(client_token["braintree_api"]["url"] + "/graphql", headers=headers, data=json.dumps(graphql_request))
        return json.loads(resp.text)

class ClientApiHttp(Http):
    def __init__(self, config, options):
        self.config = config
        self.options = options
        self.http = Http(config)

    @staticmethod
    def create():
        config = Configuration.instantiate()
        client_token = TestHelper.generate_decoded_client_token()
        authorization_fingerprint = json.loads(client_token)["authorizationFingerprint"]
        return ClientApiHttp(config, {
            "authorization_fingerprint": authorization_fingerprint,
            "shared_customer_identifier": "fake_identifier",
            "shared_customer_identifier_type": "testing"
        })

    def get(self, path):
        return self._make_request("GET", path)

    def post(self, path, params=None):
        return self._make_request("POST", path, params)

    def put(self, path, params=None):
        return self._make_request("PUT", path, params)

    def _make_request(self, http_verb, path, params=None):
        http_strategy = self.config.http_strategy()
        request_body = json.dumps(params) if params else None
        return http_strategy.http_do(http_verb, path, self.__headers(), request_body)

    def set_authorization_fingerprint(self, authorization_fingerprint):
        self.options['authorization_fingerprint'] = authorization_fingerprint

    def get_configuration(self):
        encoded_fingerprint = quote_plus(self.options["authorization_fingerprint"])
        url = "/merchants/%s/client_api/v1/configuration" % self.config.merchant_id
        url += "?authorizationFingerprint=%s" % encoded_fingerprint
        url += "&configVersion=3"

        return self.get(url)

    def get_cards(self):
        encoded_fingerprint = quote_plus(self.options["authorization_fingerprint"])
        url = "/merchants/%s/client_api/v1/payment_methods.json" % self.config.merchant_id
        url += "?authorizationFingerprint=%s" % encoded_fingerprint
        url += "&sharedCustomerIdentifier=%s" % self.options["shared_customer_identifier"]
        url += "&sharedCustomerIdentifierType=%s" % self.options["shared_customer_identifier_type"]

        return self.get(url)

    def add_card(self, params):
        url = "/merchants/%s/client_api/v1/payment_methods/credit_cards.json" % self.config.merchant_id

        if 'authorization_fingerprint' in self.options:
            params['authorizationFingerprint'] = self.options['authorization_fingerprint']

        if 'shared_customer_identifier' in self.options:
            params['sharedCustomerIdentifier'] = self.options['shared_customer_identifier']

        if 'shared_customer_identifier_type' in self.options:
            params['sharedCustomerIdentifierType'] = self.options['shared_customer_identifier_type']

        return self.post(url, params)

    def get_paypal_nonce(self, paypal_params):
        url = "/merchants/%s/client_api/v1/payment_methods/paypal_accounts" % self.config.merchant_id
        params = {"paypal_account": paypal_params}
        if 'authorization_fingerprint' in self.options:
            params['authorizationFingerprint'] = self.options['authorization_fingerprint']

        status_code, response = self.post(url, params)

        nonce = None
        if status_code == 202:
            nonce = json.loads(response)["paypalAccounts"][0]["nonce"]

        return [status_code, nonce]

    def get_credit_card_nonce(self, credit_card_params):
        url = "/merchants/%s/client_api/v1/payment_methods/credit_cards" % self.config.merchant_id
        params = {"credit_card": credit_card_params}
        if 'authorization_fingerprint' in self.options:
            params['authorizationFingerprint'] = self.options['authorization_fingerprint']

        status_code, response = self.post(url, params)

        nonce = None
        if status_code in [201, 202]:
            nonce = json.loads(response)["creditCards"][0]["nonce"]

        return [status_code, nonce]

    def __headers(self):
        return {
            "Content-type": "application/json",
            "User-Agent": "Braintree Python " + version.Version, #pylint: disable=E0602
            "X-ApiVersion": Configuration.api_version()
        }

class ExpirationHelper(Enum):
    ADYEN = "03/2030"



from typing import Dict
import braintree
from braintree.error_result import ErrorResult
from braintree.exceptions.unexpected_error import UnexpectedError
from braintree.successful_result import SuccessfulResult
from braintree.util.graphql_client import GraphQLClient
from braintree.graphql import (
    CreateCustomerSessionInput,
    UpdateCustomerSessionInput,
    CustomerRecommendations,
    CustomerRecommendationsInput,
    CustomerRecommendationsPayload,
    PaymentOptions,
)


class CustomerSessionGateway:
    """
    Creates and manages PayPal customer sessions.
    """
    def __init__(self, gateway):
        self.gateway = gateway
        self.graphql_client = self.gateway.graphql_client

    def create_customer_session(self, customer_session_input: CreateCustomerSessionInput):
        """
        Creates a new customer session.

        Example:
          customer_session = (
            CustomerSessionInput
            .builder()
            .email("test@example.com")
            .build()
          )

          create_input = (
            CreateCustomerSessionInput
            .builder()
            .customer(customer_session)
            .build()
          )

          result = gateway.customer_session.create_customer_session(create_input)
    
          if result.is_success:
            print(result.session_id)

        Args:
            CreateCustomerSessionInput: Input object for creating a customer session.
        
        Returns:
            (Successful|Error)Result: A result object with session ID if successful, or errors otherwise.

        Raises:
            UnexpectedError: If there is an unexpected error during the process.
        """
        mutation = """
            mutation CreateCustomerSession($input: CreateCustomerSessionInput!) {
              createCustomerSession(input: $input) {
                sessionId
              }
            }
        """
        variables = dict({"input": customer_session_input.to_graphql_variables()})
        return  self._execute_mutation(mutation, variables, "createCustomerSession")


    def update_customer_session(self, update_customer_session_input: UpdateCustomerSessionInput):
        """
        Updates an existing customer session.

        Example:
          customer_session = (
            CustomerSessionInput
            .builder()
            .email("test@example.com")
            .build()
          )

          update_input = (
            UpdateCustomerSessionInput
            .builder(session_id)
            .customer(customer_session)
            .build()
          )

          result = gateway.customer_session.update_customer_session(update_input)
    
          if result.is_success:
            print(result.session_id)

        Args:
            UpdateCustomerSessionInput: Input object for updating a customer session.
        
        Returns:
            (Successful|Error)Result: A result object with session ID if successful, or errors otherwise.

        Raises:
            UnexpectedError: If there is an unexpected error during the process.
        """
        mutation = """
            mutation UpdateCustomerSession($input: UpdateCustomerSessionInput!) {
              updateCustomerSession(input: $input) {
                sessionId
              }
            }
        """
        variables = dict({"input": update_customer_session_input.to_graphql_variables()})
        return self._execute_mutation(mutation, variables, "updateCustomerSession")

    def get_customer_recommendations(self, customer_recommendations_input: CustomerRecommendationsInput):
        """
        Retrieves customer recommendations associated with a customer session.

        Example:

          recommendations_input = (
            CustomerRecommendationsInput
              .builder(session_id, [Recommendations.PAYMENT_RECOMMENDATIONS])
              .build()
          )

          result = gateway.customer_session.get_customer_recommendations(recommendations_input)
         
          if result.is_success:
            print(result.customer_recommendations)

        Args:
            CustomerRecommendationsInput: Input object for retrieving customer recommendations.
        
        Returns:
            (Successful|Error)Result: A result object containing a CustomerRecommendationsPayload and a success flag if successful, or errors otherwise.

        Raises:
            UnexpectedError: If there is an unexpected error during the process.
        """
        query = """
            query CustomerRecommendations($input: CustomerRecommendationsInput!) {
                customerRecommendations(input: $input) {
                  isInPayPalNetwork
                  recommendations {
                    ... on PaymentRecommendations {
                      paymentOptions {
                        paymentOption
                        recommendedPriority
                      }
                    }
                  }
                }
              }
        """
        variables = dict({"input": customer_recommendations_input.to_graphql_variables()})

        response = self.graphql_client.query(query, variables)
        errors = GraphQLClient.get_validation_errors(response)
        if errors:
            return ErrorResult(self.gateway, {"errors": errors, "message": "Validation errors were found."})
        try:
            recommendations_payload = self._extract_customer_recommendations_payload(response)
            return SuccessfulResult({"customer_recommendations": recommendations_payload})
        except KeyError:
            raise UnexpectedError("Couldn't parse response")

    def _execute_mutation(self, mutation: str, variables: Dict, operation: str):
        response = self.graphql_client.query(mutation, variables)
        errors = GraphQLClient.get_validation_errors(response)
        if errors:
            return ErrorResult(self.gateway, {"errors": errors, "message": "Validation errors were found."})
        try:
            session_id = response["data"][operation]["sessionId"]
            return  SuccessfulResult({"session_id": session_id})
        except KeyError:
            raise UnexpectedError("Couldn't parse response")

    def _extract_customer_recommendations_payload(self, response):
        customer_recommendations = response["data"]["customerRecommendations"]

        is_in_paypal_network = customer_recommendations["isInPayPalNetwork"]
        recommendations = [] if customer_recommendations["recommendations"] is None else customer_recommendations["recommendations"]["paymentOptions"]
        payment_options = [
            PaymentOptions(
                    recommendation_data["paymentOption"],
                    recommendation_data["recommendedPriority"]
            ) 
            for recommendation_data in recommendations
        ]

        recommendations_union = CustomerRecommendations(payment_options)
        return CustomerRecommendationsPayload(is_in_paypal_network, recommendations_union)






from braintree.credit_card import CreditCard
from braintree.search import Search
from braintree.transaction import Transaction
from braintree.util import Constants

class TransactionSearch:
    billing_first_name           = Search.TextNodeBuilder("billing_first_name")
    billing_company              = Search.TextNodeBuilder("billing_company")
    billing_country_name         = Search.TextNodeBuilder("billing_country_name")
    billing_extended_address     = Search.TextNodeBuilder("billing_extended_address")
    billing_first_name           = Search.TextNodeBuilder("billing_first_name")
    billing_last_name            = Search.TextNodeBuilder("billing_last_name")
    billing_locality             = Search.TextNodeBuilder("billing_locality")
    billing_postal_code          = Search.TextNodeBuilder("billing_postal_code")
    billing_region               = Search.TextNodeBuilder("billing_region")
    billing_street_address       = Search.TextNodeBuilder("billing_street_address")
    credit_card_cardholder_name  = Search.TextNodeBuilder("credit_card_cardholder_name")
    currency                     = Search.TextNodeBuilder("currency")
    customer_company             = Search.TextNodeBuilder("customer_company")
    customer_email               = Search.TextNodeBuilder("customer_email")
    customer_fax                 = Search.TextNodeBuilder("customer_fax")
    customer_first_name          = Search.TextNodeBuilder("customer_first_name")
    customer_id                  = Search.TextNodeBuilder("customer_id")
    customer_last_name           = Search.TextNodeBuilder("customer_last_name")
    customer_phone               = Search.TextNodeBuilder("customer_phone")
    customer_website             = Search.TextNodeBuilder("customer_website")
    id                           = Search.TextNodeBuilder("id")
    order_id                     = Search.TextNodeBuilder("order_id")
    payment_method_token         = Search.TextNodeBuilder("payment_method_token")
    processor_authorization_code = Search.TextNodeBuilder("processor_authorization_code")
    europe_bank_account_iban       = Search.TextNodeBuilder("europe_bank_account_iban")
    settlement_batch_id          = Search.TextNodeBuilder("settlement_batch_id")
    shipping_company             = Search.TextNodeBuilder("shipping_company")
    shipping_country_name        = Search.TextNodeBuilder("shipping_country_name")
    shipping_extended_address    = Search.TextNodeBuilder("shipping_extended_address")
    shipping_first_name          = Search.TextNodeBuilder("shipping_first_name")
    shipping_last_name           = Search.TextNodeBuilder("shipping_last_name")
    shipping_locality            = Search.TextNodeBuilder("shipping_locality")
    shipping_postal_code         = Search.TextNodeBuilder("shipping_postal_code")
    shipping_region              = Search.TextNodeBuilder("shipping_region")
    shipping_street_address      = Search.TextNodeBuilder("shipping_street_address")
    paypal_payer_email           = Search.TextNodeBuilder("paypal_payer_email")
    paypal_payment_id            = Search.TextNodeBuilder("paypal_payment_id")
    paypal_authorization_id      = Search.TextNodeBuilder("paypal_authorization_id")
    sepa_debit_paypal_v2_order_id = Search.TextNodeBuilder("sepa_debit_paypal_v2_order_id")
    credit_card_unique_identifier = Search.TextNodeBuilder("credit_card_unique_identifier")
    store_id                     = Search.TextNodeBuilder("store_id")

    credit_card_expiration_date  = Search.EqualityNodeBuilder("credit_card_expiration_date")
    credit_card_number           = Search.PartialMatchNodeBuilder("credit_card_number")
    
    user                         = Search.MultipleValueNodeBuilder("user")
    ids                          = Search.MultipleValueNodeBuilder("ids")
    merchant_account_id          = Search.MultipleValueNodeBuilder("merchant_account_id")
    payment_instrument_type      = Search.MultipleValueNodeBuilder("payment_instrument_type")
    store_ids                    = Search.MultipleValueNodeBuilder("store_ids")

    created_using = Search.MultipleValueNodeBuilder(
        "created_using",
        Constants.get_all_constant_values_from_class(Transaction.CreatedUsing)
    )

    credit_card_card_type = Search.MultipleValueNodeBuilder(
        "credit_card_card_type",
        Constants.get_all_constant_values_from_class(CreditCard.CardType)
    )

    credit_card_customer_location = Search.MultipleValueNodeBuilder(
        "credit_card_customer_location",
        Constants.get_all_constant_values_from_class(CreditCard.CustomerLocation)
    )

    debit_network = Search.MultipleValueNodeBuilder(
        "debit_network",
        Constants.get_all_enum_values(CreditCard.DebitNetwork)
    )

    source = Search.MultipleValueNodeBuilder("source")

    status = Search.MultipleValueNodeBuilder(
        "status",
        Constants.get_all_constant_values_from_class(Transaction.Status)
    )

    type = Search.MultipleValueNodeBuilder(
        "type",
        Constants.get_all_constant_values_from_class(Transaction.Type)
    )

    refund = Search.KeyValueNodeBuilder("refund")

    amount = Search.RangeNodeBuilder("amount")
    authorization_expired_at = Search.RangeNodeBuilder("authorization_expired_at")
    authorized_at = Search.RangeNodeBuilder("authorized_at")
    created_at = Search.RangeNodeBuilder("created_at")
    disbursement_date = Search.RangeNodeBuilder("disbursement_date")
    dispute_date = Search.RangeNodeBuilder("dispute_date")
    failed_at = Search.RangeNodeBuilder("failed_at")
    gateway_rejected_at = Search.RangeNodeBuilder("gateway_rejected_at")
    processor_declined_at = Search.RangeNodeBuilder("processor_declined_at")
    settled_at = Search.RangeNodeBuilder("settled_at")
    submitted_for_settlement_at = Search.RangeNodeBuilder("submitted_for_settlement_at")
    voided_at = Search.RangeNodeBuilder("voided_at")
    ach_return_responses_created_at = Search.RangeNodeBuilder("ach_return_responses_created_at")
    reason_code = Search.MultipleValueNodeBuilder('reason_code')



import braintree

class PaginatedCollection(object):
    """
    A class representing results from a paginated list. Supports the iterator protocol::

        results = braintree.MerchantAccount.all()
        for merchant_account in results.items:
            print merchant_account.id
    """

    def __init__(self, method):
        self.__method = method

    @property
    def items(self):
        """ Returns a generator allowing iteration over all of the results. """
        current_page = 0
        total_items = 0
        while True:
            current_page += 1

            results = self.__method(current_page)
            total_items = results.total_items

            for item in results.current_page:
                yield item

            if current_page * results.page_size >= total_items:
                break

    def __iter__(self):
        return self.items



import braintree
from braintree.resource import Resource

# NEXT_MAJOR_VERSION - rename to GooglePayCard
class AndroidPayCard(Resource):
    """
    A class representing Braintree Android Pay card objects.
    """
    def __init__(self, gateway, attributes):
        Resource.__init__(self, gateway, attributes)
        if hasattr(self, 'expired'):
            self.is_expired = self.expired

        if "subscriptions" in attributes:
            self.subscriptions = [braintree.subscription.Subscription(gateway, subscription) for subscription in self.subscriptions]

    @property
    def expiration_date(self):
        if not self.expiration_month or not self.expiration_year:
            return None
        return self.expiration_month + "/" + self.expiration_year

    @property
    def last_4(self):
        return self.virtual_card_last_4

    @property
    def card_type(self):
        return self.virtual_card_type

    @staticmethod
    def signature():
        options = ["make_default"]

        signature = [
            "customer_id",
            "cardholder_name",
            "cryptogram",
            "google_transaction_id",
            "expiration_month",
            "expiration_year",
            "number",
            "token",
            "eci_indicator",
            {
                "options": options
            },
            {
                "billing_address": [
                    "company",
                    "country_code_alpha2",
                    "country_code_alpha3",
                    "country_code_numeric",
                    "country_name",
                    "extended_address",
                    "first_name",
                    "last_name",
                    "locality",
                    "postal_code",
                    "region",
                    "street_address",
                    "phone_number"
                ]
            }
        ]

        return signature


    @staticmethod
    def card_signature():
        options = ["make_default"]

        signature = [
            "customer_id",
            "cardholder_name",
            "google_transaction_id",
            "expiration_month",
            "expiration_year",
            "number",
            "token",
            {
                "options": options
            },
            {
                "billing_address": [
                    "company",
                    "country_code_alpha2",
                    "country_code_alpha3",
                    "country_code_numeric",
                    "country_name",
                    "extended_address",
                    "first_name",
                    "last_name",
                    "locality",
                    "postal_code",
                    "region",
                    "street_address",
                    "phone_number"
                ]
            }
        ]

        return signature


    @staticmethod
    def network_token_signature():
        return AndroidPayCard.signature()



from braintree.util.http import Http
import braintree
from braintree.add_on import AddOn
from braintree.configuration import Configuration
from braintree.discount import Discount
from braintree.resource_collection import ResourceCollection
from braintree.resource import Resource
from braintree.successful_result import SuccessfulResult
from braintree.error_result import ErrorResult

class Plan(Resource):

    def __init__(self, gateway, attributes):
        Resource.__init__(self, gateway, attributes)
        if "add_ons" in attributes:
            self.add_ons = [AddOn(gateway, add_on) for add_on in self.add_ons]
        if "discounts" in attributes:
            self.discounts = [Discount(gateway, discount) for discount in self.discounts]

    @staticmethod
    def all():
        return Configuration.gateway().plan.all()

    @staticmethod
    def create(params=None):
        if params is None:
            params = {}
        return Configuration.gateway().plan.create(params)

    @staticmethod
    def find(subscription_id):
        return Configuration.gateway().plan.find(subscription_id)

    @staticmethod
    def update(subscription_id, params=None):
        if params is None:
            params = {}
        return Configuration.gateway().plan.update(subscription_id, params)

    @staticmethod
    def create_signature():
        return [
            "billing_day_of_month",
            "billing_frequency",
            "currency_iso_code",
            "description",
            "id",
            "merchant_id",
            "name",
            "number_of_billing_cycles",
            "price",
            "trial_duration",
            "trial_duration_unit",
            "trial_period"
        ] + Plan._add_on_discount_signature()

    @staticmethod
    def update_signature():
        return [
            "billing_day_of_month",
            "billing_frequency",
            "currency_iso_code",
            "description",
            "id",
            "merchant_id",
            "name",
            "number_of_billing_cycles",
            "price",
            "trial_duration",
            "trial_duration_unit",
            "trial_period"
        ] + Plan._add_on_discount_signature()

    @staticmethod
    def _add_on_discount_signature():
        return [
            {
                "add_ons": [
                    {"add": ["amount", "inherited_from_id", "never_expires", "number_of_billing_cycles", "quantity"]},
                    {"update": ["amount", "existing_id", "never_expires", "number_of_billing_cycles", "quantity"]},
                    {"remove": ["_any_key_"]}
                ]
            },
            {
                "discounts": [
                    {"add": ["amount", "inherited_from_id", "never_expires", "number_of_billing_cycles", "quantity"]},
                    {"update": ["amount", "existing_id", "never_expires", "number_of_billing_cycles", "quantity"]},
                    {"remove": ["_any_key_"]}
                ]
            }
        ]




from braintree.exchange_rate_quote_payload import ExchangeRateQuotePayload
from braintree.error_result import ErrorResult
from braintree.successful_result import SuccessfulResult

class ExchangeRateQuoteGateway(object):
    def __init__(self, gateway, graphql_client = None):
        self.gateway = gateway
        self.config = gateway.config
        self.graphql_client = None if graphql_client is None else graphql_client
      
    def generate(self, request):
        definition = """
          mutation ($exchangeRateQuoteRequest: GenerateExchangeRateQuoteInput!) {
            generateExchangeRateQuote(input: $exchangeRateQuoteRequest) {
              quotes {
                id
                baseAmount {value, currencyCode}
                quoteAmount {value, currencyCode}
                exchangeRate
                tradeRate
                expiresAt
                refreshesAt
              }
            }
          }"""

        param = request.to_graphql_variables()
        graphql_client = self.graphql_client if self.graphql_client is not None else self.gateway.graphql_client
        response = graphql_client.query(definition, param)

        if "data" in response and "generateExchangeRateQuote" in response["data"]:
            result = response["data"]["generateExchangeRateQuote"]
            self.exchange_rate_quote_payload = ExchangeRateQuotePayload(result)
            return SuccessfulResult({"exchange_rate_quote_payload": self.exchange_rate_quote_payload})
        elif "errors" in response:
            error_codes = response["errors"][0]
            error_codes["errors"] = dict()
            return ErrorResult(self.gateway, error_codes)


class Nonces(object):
    AbstractTransactable = "fake-abstract-transactable-nonce"
    # NEXT_MAJOR_VERSION remove amex express checkout
    AmexExpressCheckoutCard = "fake-amex-express-checkout-nonce"
    # NEXT_MAJOR_VERSION - rename AndroidPay to GooglePay
    AndroidPayCard = "fake-android-pay-nonce"
    AndroidPayCardAmEx = "fake-android-pay-amex-nonce"
    AndroidPayCardDiscover = "fake-android-pay-discover-nonce"
    AndroidPayCardMasterCard = "fake-android-pay-mastercard-nonce"
    AndroidPayCardVisa = "fake-android-pay-visa-nonce"
    ApplePayAmEx = "fake-apple-pay-amex-nonce"
    ApplePayMasterCard = "fake-apple-pay-mastercard-nonce"
    ApplePayMpan = "fake-apple-pay-mpan-nonce"
    ApplePayVisa = "fake-apple-pay-visa-nonce"
    Consumed = "fake-consumed-nonce"
    Europe = "fake-europe-bank-account-nonce"
    GatewayRejectedFraud = "fake-gateway-rejected-fraud-nonce"
    GatewayRejectedRiskThreshold = "fake-gateway-rejected-risk-thresholds-nonce"
    LocalPayment = "fake-local-payment-method-nonce"
    LuhnInvalid = "fake-luhn-invalid-nonce"
    # NEXT_MAJOR_VERSION remove masterpass
    MasterpassAmEx = "fake-masterpass-amex-nonce"
    MasterpassDiscover = "fake-masterpass-discover-nonce"
    MasterpassMasterCard = "fake-masterpass-mastercard-nonce"
    MasterpassVisa = "fake-masterpass-visa-nonce"
    PayPalBillingAgreement = "fake-paypal-billing-agreement-nonce"
    # NEXT_MAJOR_VERSION - no longer supported in the Gateway remove this constant
    PayPalFuturePayment = "fake-paypal-future-nonce"
    PayPalFuturePaymentRefreshToken = "fake-paypal-future-refresh-token-nonce"
    PayPalOneTimePayment = "fake-paypal-one-time-nonce"
    ProcessorDeclinedAmEx = "fake-processor-declined-amex-nonce"
    ProcessorDeclinedDiscover = "fake-processor-declined-discover-nonce"
    ProcessorDeclinedMasterCard = "fake-processor-declined-mastercard-nonce"
    ProcessorDeclinedVisa = "fake-processor-declined-visa-nonce"
    ProcessorFailureJCB = "fake-processor-failure-jcb-nonce"
    SEPA = "fake-sepa-bank-account-nonce"
    MetaCheckoutCard = "fake-meta-checkout-card-nonce"
    MetaCheckoutToken = "fake-meta-checkout-token-nonce"
    VisaCheckoutAmEx = "fake-visa-checkout-amex-nonce"
    VisaCheckoutDiscover = "fake-visa-checkout-discover-nonce"
    VisaCheckoutMasterCard = "fake-visa-checkout-mastercard-nonce"
    VisaCheckoutVisa = "fake-visa-checkout-visa-nonce"
    # NEXT_MAJOR_VERSION remove SamsungPay nonces
    SamsungPayAmex = "tokensam_fake_american_express"
    SamsungPayDiscover = "tokensam_fake_american_express"
    SamsungPayMasterCard = "tokensam_fake_mastercard"
    SamsungPayVisa = "tokensam_fake_visa"
    SepaDirectDebit = "fake-sepa-direct-debit-nonce"
    ThreeDSecureTwoVisaErrorOnLookup = "fake-three-d-secure-two-visa-error-on-lookup-nonce"
    ThreeDSecureTwoVisaSuccessfulFrictionlessAuthentication = "fake-three-d-secure-two-visa-successful-frictionless-authentication-nonce"
    ThreeDSecureTwoVisaSuccessfulStepUpAuthentication = "fake-three-d-secure-two-visa-successful-step-up-authentication-nonce"
    ThreeDSecureTwoVisaTimeoutOnLookup = "fake-three-d-secure-two-visa-timeout-on-lookup-nonce"
    ThreeDSecureVisaAttemptsNonParticipating = "fake-three-d-secure-visa-attempts-non-participating-nonce"
    ThreeDSecureVisaAuthenticationUnavailable = "fake-three-d-secure-visa-authentication-unavailable-nonce"
    ThreeDSecureVisaBypassedAuthentication = "fake-three-d-secure-visa-bypassed-authentication-nonce"
    ThreeDSecureVisaFailedAuthentication = "fake-three-d-secure-visa-failed-authentication-nonce"
    ThreeDSecureVisaFailedSignature = "fake-three-d-secure-visa-failed-signature-nonce"
    ThreeDSecureVisaFullAuthentication = "fake-three-d-secure-visa-full-authentication-nonce"
    ThreeDSecureVisaLookupTimeout = "fake-three-d-secure-visa-lookup-timeout-nonce"
    ThreeDSecureVisaMPIAuthenticateError = "fake-three-d-secure-visa-mpi-authenticate-error-nonce"
    ThreeDSecureVisaMPILookupError = "fake-three-d-secure-visa-mpi-lookup-error-nonce"
    ThreeDSecureVisaNoteEnrolled = "fake-three-d-secure-visa-not-enrolled-nonce"
    ThreeDSecureVisaUnavailable = "fake-three-d-secure-visa-unavailable-nonce"
    Transactable = "fake-valid-nonce"
    TransactableAmEx = "fake-valid-amex-nonce"
    TransactableCommercial = "fake-valid-commercial-nonce"
    TransactableCountryOfIssuanceCAD = "fake-valid-country-of-issuance-cad-nonce"
    TransactableCountryOfIssuanceUSA = "fake-valid-country-of-issuance-usa-nonce"
    TransactableDebit = "fake-valid-debit-nonce"
    TransactableDinersClub = "fake-valid-dinersclub-nonce"
    TransactableDiscover = "fake-valid-discover-nonce"
    TransactableDurbinRegulated = "fake-valid-durbin-regulated-nonce"
    TransactableHealthcare = "fake-valid-healthcare-nonce"
    TransactableIssuingBankNetworkOnly = "fake-valid-issuing-bank-network-only-nonce"
    TransactableJCB = "fake-valid-jcb-nonce"
    TransactableMaestro = "fake-valid-maestro-nonce"
    TransactableMasterCard = "fake-valid-mastercard-nonce"
    TransactableNoIndicators = "fake-valid-no-indicators-nonce"
    TransactablePayroll = "fake-valid-payroll-nonce"
    TransactablePinlessDebitVisa = "fake-pinless-debit-visa-nonce"
    TransactablePrepaid = "fake-valid-prepaid-nonce"
    TransactableUnknownIndicators = "fake-valid-unknown-indicators-nonce"
    TransactableVisa = "fake-valid-visa-nonce"
    VenmoAccount = "fake-venmo-account-nonce"
    VenmoAccountTokenIssuanceError = "fake-token-issuance-error-venmo-account-nonce"
    VisaCheckoutAmEx = "fake-visa-checkout-amex-nonce"
    VisaCheckoutDiscover = "fake-visa-checkout-discover-nonce"
    VisaCheckoutMasterCard = "fake-visa-checkout-mastercard-nonce"
    VisaCheckoutVisa = "fake-visa-checkout-visa-nonce"
    UsBankAccount = "fake-us-bank-account-nonce"


import braintree

from braintree.resource import Resource

class BlikAlias(Resource):
    """
    A class representing a BlikAlias.

    For more information on BlikAliases, see https://developer.paypal.com/braintree/docs/guides/local-payment-methods/blik-one-click

    """

    def __init__(self, gateway, attributes):
        Resource.__init__(self, gateway, attributes)




import braintree
from braintree.resource import Resource

class UnknownPaymentMethod(Resource):
    def image_url(self):
        return "https://assets.braintreegateway.com/payment_method_logo/unknown.png"



import json

from braintree.exceptions.authentication_error import AuthenticationError
from braintree.exceptions.authorization_error import AuthorizationError
from braintree.exceptions.service_unavailable_error import ServiceUnavailableError
from braintree.exceptions.not_found_error import NotFoundError
from braintree.exceptions.server_error import ServerError
from braintree.exceptions.too_many_requests_error import TooManyRequestsError
from braintree.exceptions.unexpected_error import UnexpectedError
from braintree.exceptions.upgrade_required_error import UpgradeRequiredError
from braintree.util.http import Http


class GraphQLClient(Http):

    @staticmethod
    def raise_exception_for_graphql_error(response):
        if "errors" not in response:
            return

        for error in response["errors"]:
            if "extensions" in error and "errorClass" in error["extensions"]:
                error_type = error["extensions"]["errorClass"]
                if error_type == "VALIDATION":
                    continue
                elif error_type == "AUTHENTICATION":
                    raise AuthenticationError(error["message"])
                elif error_type == "AUTHORIZATION":
                    raise AuthorizationError(error["message"])
                elif error_type == "NOT_FOUND":
                    raise NotFoundError
                elif error_type == "UNSUPPORTED_CLIENT":
                    raise UpgradeRequiredError("Please upgrade your client library.")
                elif error_type == "RESOURCE_LIMIT":
                    raise TooManyRequestsError
                elif error_type == "INTERNAL":
                    raise ServerError
                elif error_type == "SERVICE_AVAILABILITY":
                    raise ServiceUnavailableError
                else:
                    raise UnexpectedError("Unexpected Response: " + error["message"])

    def __init__(self, config=None, environment=None):
        Http.__init__(self, config, environment)
        self.graphql_headers = {
            "Accept": "application/json",
            "Braintree-Version": config.graphql_api_version(),
            "Content-Type": "application/json",
        }

    def query(self, definition, variables=None, operation_name=None):
        graphql_request = {"query": definition}

        if variables is not None:
            graphql_request["variables"] = variables

        if operation_name is not None:
            graphql_request["operationName"] = operation_name

        response = self._make_request(
            "POST",
            self.config.graphql_base_url(),
            Http.ContentType.Json,
            json.dumps(graphql_request),
            header_overrides=self.graphql_headers,
        )
        self.raise_exception_for_graphql_error(response)

        return response

    @staticmethod
    def get_validation_errors(response):
        if "errors" not in response or not isinstance(response["errors"], list):
            return None

        validation_errors = [
            {
                "attribute": "",
                "code": GraphQLClient.get_validation_error_code(error),
                "message": error["message"],
            }
            for error in response["errors"]
        ]
        return {"errors": validation_errors}

    @staticmethod
    def get_validation_error_code(error):
        try:
            return error["extensions"]["legacyCode"]
        except KeyError:
            return None



import mimetypes
from braintree.successful_result import SuccessfulResult
from braintree.resource import Resource
from braintree.configuration import Configuration


class DocumentUpload(Resource):
    """
    A class representing a DocumentUpload.

    An example of creating a document upload with all available fields:

        result = braintree.DocumentUpload.create(
            {
                "kind": braintree.DocumentUpload.Kind.EvidenceDocument,
                "file": open("path/to/file", "rb"),
            }
        )

    For more information on DocumentUploads, see https://developer.paypal.com/braintree/docs/reference/request/document-upload/create

    """

    class Kind(object):
        EvidenceDocument = "evidence_document"

    @staticmethod
    def create(params=None):
        """
        Create a DocumentUpload

        File and Kind are required:

            result = braintree.DocumentUpload.create(
                {
                    "kind": braintree.DocumentUpload.Kind.EvidenceDocument,
                    "file": open("path/to/file", "rb"),
                }
            )

        """
        if params is None:
            params = {}
        return Configuration.gateway().document_upload.create(params)

    @staticmethod
    def create_signature():
        return [
            "kind",
            "file",
        ]

    def __init__(self, gateway, attributes):
        Resource.__init__(self, gateway, attributes)



import braintree
from braintree.errors import Errors
from braintree.credit_card_verification import CreditCardVerification

class ErrorResult(object):
    """
    An instance of this class is returned from most operations when there is a validation error.  Call :func:`errors` to get the collection of errors::

        error_result = Transaction.sale({})
        assert(error_result.is_success == False)
        assert(error_result.errors.for_object("transaction").on("amount")[0].code == ErrorCodes.Transaction.AmountIsRequired)

    Errors can be nested at different levels.  For example, creating a transaction with a credit card can have errors at the transaction level as well as the credit card level.  :func:`for_object` returns the :class:`ValidationErrorCollection <braintree.validation_error_collection.ValidationErrorCollection>` for the errors at that level.  For example::

        error_result = Transaction.sale({"credit_card": {"number": "invalid"}})
        assert(error_result.errors.for_object("transaction").for_object("credit_card").on("number")[0].code == ErrorCodes.CreditCard.NumberHasInvalidLength)
    """

    def __init__(self, gateway, attributes):
        if "params" in attributes:
            self.params = attributes["params"]
        else:
            self.params = None

        self.errors = Errors(attributes["errors"])
        self.message = attributes["message"]

        if "verification" in attributes:
            self.credit_card_verification = CreditCardVerification(gateway, attributes["verification"])
        else:
            self.credit_card_verification = None

        if "transaction" in attributes:
            self.transaction = braintree.transaction.Transaction(gateway, attributes["transaction"])
        else:
            self.transaction = None

        if "subscription" in attributes:
            self.subscription = braintree.subscription.Subscription(gateway, attributes["subscription"])
        else:
            self.subscription = None

        if "plan" in attributes:
            self.plan = braintree.plan.Plan(gateway, attributes["plan"])
        else:
            self.plan = None

        if "merchant_account" in attributes:
            self.merchant_account = braintree.merchant_account.MerchantAccount(gateway, attributes["merchant_account"])
        else:
            self.merchant_account = None

    def __repr__(self):
        return "<%s '%s' at %x>" % (self.__class__.__name__, self.message, id(self))

    @property
    def is_success(self):
        """ Returns whether the result from the gateway is a successful response. """

        return False



import datetime
import sys
from decimal import Decimal

integer_types = int
text_type = str
binary_type = bytes

class Generator(object):
    def __init__(self, dict):
        self.dict = dict

    def generate(self):
        return self.__generate_dict(self.dict)

    def __escape(self, value):
        return value.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;").replace("'", "&apos;").replace('"', "&quot;")

    def __generate_boolean(self, value):
        return str(value).lower()

    def __generate_datetime(self, value):
        return value.strftime("%Y-%m-%dT%H:%M:%SZ")

    def __generate_dict(self, dictionary):
        xml = ""
        for key, val in dictionary.items():
            xml += self.__generate_node(key, val)
        return xml

    def __generate_list(self, list):
        xml = ""
        for item in list:
            xml += self.__generate_node("item", item)
        return xml

    def __generate_node(self, key, value):
        open_tag = "<" + self.__escape(key) + ">"
        close_tag = "</" + self.__escape(key) + ">"

        if isinstance(value, text_type):
            return open_tag + self.__escape(value).encode('ascii', 'xmlcharrefreplace').decode('utf-8') + close_tag
        elif isinstance(value, binary_type):
            return open_tag + self.__escape(value) + close_tag
        elif isinstance(value, Decimal):
            return open_tag + str(value) + close_tag
        elif isinstance(value, dict):
            return open_tag + self.__generate_dict(value) + close_tag
        elif isinstance(value, list):
            open_tag = "<" + self.__escape(key) + " type=\"array\">"
            return open_tag + self.__generate_list(value) + close_tag
        elif isinstance(value, bool):
            open_tag = "<" + self.__escape(key) + " type=\"boolean\">"
            return open_tag + self.__generate_boolean(value) + close_tag
        elif isinstance(value, integer_types) and not isinstance(value, bool):
            open_tag = "<" + self.__escape(key) + " type=\"integer\">"
            return open_tag + str(value) + close_tag
        elif isinstance(value, type(None)):
            return open_tag + close_tag
        elif isinstance(value, datetime.datetime) or isinstance(value, datetime.date):
            open_tag = "<" + self.__escape(key) + " type=\"datetime\">"
            return open_tag + self.__generate_datetime(value) + close_tag
        else:
            raise RuntimeError("Unexpected XML node type: " + str(type(value)))



import braintree
from braintree.resource import Resource
from braintree.client_token import ClientToken
from braintree import exceptions


class ClientTokenGateway(object):
    def __init__(self, gateway):
        self.gateway = gateway
        self.config = gateway.config

    def generate(self, params=None):
        if params is None:
            params = {}
        if "options" in params and "customer_id" not in params:
            for option in ["fail_on_duplicate_payment_method", "fail_on_duplicate_payment_method_for_customer", "make_default", "verify_card"]:
                if option in params["options"]:
                    raise exceptions.InvalidSignatureError("cannot specify %s without a customer_id" % option)

        if "version" not in params:
            params["version"] = 2

        Resource.verify_keys(params, ClientToken.generate_signature())
        params = {'client_token': params}

        response = self.config.http().post(self.config.base_merchant_path() + "/client_token", params)

        if "client_token" in response:
            return response["client_token"]["value"]
        else:
            raise ValueError(response["api_error_response"]["message"])



from braintree.validation_error import ValidationError


class ValidationErrorCollection(object):
    """
    A class representing a collection of validation errors.

    For more information on ValidationErrors, see https://developer.paypal.com/braintree/docs/reference/general/validation-errors/overview/python

    """

    def __init__(self, data=None):
        if data is None:
            data = {"errors": []}
        self.data = data

    @property
    def deep_errors(self):
        """
        Return all :class:`ValidationErrors <braintree.validation_error.ValidationError>`, including nested errors.
        """

        result = []
        result.extend(self.errors)
        for nested_error in self.__nested_errors.values():
            result.extend(nested_error.deep_errors)
        return result

    def for_index(self, index):
        return self.for_object("index_%s" % index)

    def for_object(self, nested_key):
        """
        Returns a :class:`ValidationErrorCollection <braintree.validation_error_collection.ValidationErrorCollection>`

        It represents the errors at the nested level:::

            error_result = Transaction.sale({"credit_card": {"number": "invalid"}})
            print error_result.errors.for_object("transaction").for_object("credit_card").on("number")[0].code

        """

        return self.__get_nested_errors(nested_key)

    def on(self, attribute):
        """
        Returns the list of errors

        Restricted to a given attribute::

            error_result = Transaction.sale({"credit_card": {"number": "invalid"}})
            print [ error.code for error in error_result.errors.for_object("transaction").for_object("credit_card").on("number") ]

        """
        return [error for error in self.errors if error.attribute == attribute]

    @property
    def deep_size(self):
        """Returns the number of errors on this object and any nested objects."""

        size = len(self.errors)
        for error in self.__nested_errors.values():
            size += error.deep_size
        return size

    @property
    def errors(self):
        """Returns a list of :class:`ValidationError <braintree.validation_error.ValidationError>` objects."""

        return [ValidationError(error) for error in self.data["errors"]]

    @property
    def size(self):
        """Returns the number of errors on this object, without counting nested errors."""
        return len(self.errors)

    def __get_nested_errors(self, nested_key):
        if nested_key in self.__nested_errors:
            return self.__nested_errors[nested_key]
        else:
            return ValidationErrorCollection()

    def __getitem__(self, index):
        return self.errors[index]

    def __len__(self):
        return self.size

    @property
    def __nested_errors(self):
        nested_errors = {}
        for key in self.data:
            if key == "errors":
                continue
            nested_errors[key] = ValidationErrorCollection(self.data[key])
        return nested_errors




class PaymentInstrumentType():
    # NEXT_MAJOR_VERSION remove amex express checkout, masterpass, 
    # and SamsungPayCard. They have been deprecated
    AmexExpressCheckoutCard = "amex_express_checkout_card"
    AndroidPayCard = "android_pay_card"
    ApplePayCard = "apple_pay_card"
    CreditCard = "credit_card"
    EuropeBankAccount = "europe_bank_account"
    LocalPayment = "local_payment"
    MasterpassCard = "masterpass_card"
    MetaCheckoutCard = "meta_checkout_card"
    MetaCheckoutToken = "meta_checkout_token"
    PayPalAccount = "paypal_account"
    PayPalHere = "paypal_here"
    SamsungPayCard = "samsung_pay_card"
    SepaDirectDebitAccount = "sepa_debit_account"
    UsBankAccount = "us_bank_account"
    VenmoAccount = "venmo_account"
    VisaCheckoutCard = "visa_checkout_card"



from braintree.attribute_getter import AttributeGetter
from braintree.liability_shift import LiabilityShift

class RiskData(AttributeGetter):
    def __init__(self, attributes):
        AttributeGetter.__init__(self, attributes)
        if "liability_shift" in attributes:
            self.liability_shift = LiabilityShift(attributes["liability_shift"])



import braintree
from braintree.address import Address
from braintree.resource import Resource

# NEXT_MAJOR_VERSION remove this class
# SamsungPay is deprecated
class SamsungPayCard(Resource):
    def __init__(self, gateway, attributes):
        Resource.__init__(self, gateway, attributes)

        if "billing_address" in attributes:
            self.billing_address = Address(gateway, self.billing_address)
        else:
            self.billing_address = None

        if "subscriptions" in attributes:
            self.subscriptions = [braintree.subscription.Subscription(gateway, subscription) for subscription in self.subscriptions]

    @property
    def expiration_date(self):
        if not self.expiration_month or not self.expiration_year:
            return None
        return self.expiration_month + "/" + self.expiration_year

    @property
    def masked_number(self):
        return self.bin + "******" + self.last_4



from html import escape
from braintree.apple_pay_options import ApplePayOptions
from braintree.error_result import ErrorResult
from braintree.successful_result import SuccessfulResult
from braintree.exceptions.unexpected_error import UnexpectedError

class ApplePayGateway(object):
    def __init__(self, gateway):
        self.gateway = gateway
        self.config = gateway.config

    def register_domain(self, domain):
        response = self.config.http().post(self.config.base_merchant_path() + "/processing/apple_pay/validate_domains", {'url': domain})

        if "response" in response and response["response"]["success"]:
            return SuccessfulResult()
        elif response["api_error_response"]:
            return ErrorResult(self.gateway, response["api_error_response"])

    def unregister_domain(self, domain):
        self.config.http().delete(self.config.base_merchant_path() + "/processing/apple_pay/unregister_domain?url=" + escape(domain))
        return SuccessfulResult()

    def registered_domains(self):
        response = self.config.http().get(self.config.base_merchant_path() + "/processing/apple_pay/registered_domains")

        if "response" in response:
            response = ApplePayOptions(response.pop("response"))

        return response.domains



from braintree.exceptions.braintree_error import BraintreeError

class UpgradeRequiredError(BraintreeError):
    """
    Raised for unsupported client library versions.

    See https://developer.paypal.com/braintree/docs/reference/general/exceptions/python#upgrade-required-error
    """
    pass



import braintree
from braintree.resource import Resource
from braintree.configuration import Configuration


class SepaDirectDebitAccount(Resource):
    @staticmethod
    def find(sepa_direct_debit_account_token):
        return Configuration.gateway().sepa_direct_debit_account.find(sepa_direct_debit_account_token)

    @staticmethod
    def delete(sepa_direct_debit_account_token):
        return Configuration.gateway().sepa_direct_debit_account.delete(sepa_direct_debit_account_token)

    def __init__(self, gateway, attributes):
        Resource.__init__(self, gateway, attributes)
        if "subscriptions" in attributes:
            self.subscriptions = [braintree.subscription.Subscription(gateway, subscription) for subscription in self.subscriptions]



import braintree
from braintree.sepa_direct_debit_account import SepaDirectDebitAccount
from braintree.error_result import ErrorResult
from braintree.exceptions.not_found_error import NotFoundError
from braintree.resource import Resource
from braintree.successful_result import SuccessfulResult


class SepaDirectDebitAccountGateway(object):
    def __init__(self, gateway):
        self.gateway = gateway
        self.config = gateway.config

    def find(self, sepa_direct_debit_account_token):
        try:
            if sepa_direct_debit_account_token is None or sepa_direct_debit_account_token.strip() == "":
                raise NotFoundError()

            response = self.config.http().get(self.config.base_merchant_path() + "/payment_methods/sepa_debit_account/" + sepa_direct_debit_account_token)
            if "sepa_debit_account" in response:
                return SepaDirectDebitAccount(self.gateway, response["sepa_debit_account"])
        except NotFoundError:
            raise NotFoundError("sepa direct debit account with token " + repr(sepa_direct_debit_account_token) + " not found")

    def delete(self, sepa_direct_debit_account_token):
        self.config.http().delete(self.config.base_merchant_path() + "/payment_methods/sepa_debit_account/" + sepa_direct_debit_account_token)
        return SuccessfulResult()
